@@ -1,6 +1,7 @@
 #define __GNOME_FONT_FACE_C__
 
 #include <config.h>
+#include <stdlib.h>
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -8,10 +9,17 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <stdio.h>
+#include <ctype.h>
 #include <freetype/ftoutln.h>
+#include <locale.h>
+
 #include <libgnomeprint/gnome-print-i18n.h>
 #include <libgnomeprint/gnome-font.h>
 #include <libgnomeprint/gnome-font-private.h>
+
+#define USE_GNOME_FONT_FACE_EXPERIMENTAL_INTERFACE
+#include <libgnomeprint/gnome-font-face-private.h>
+
 #include <libgnomeprint/gp-unicode.h>
 #include <libgnomeprint/gp-ps-unicode.h>
 #include "gp-truetype-utils.h"
@@ -57,6 +65,7 @@ static void gff_face_from_entry (GPFontEntry *e);
 
 #define GFE_IS_T1_A(e) (((e)->type == GP_FONT_ENTRY_TYPE1) || ((e)->type == GP_FONT_ENTRY_TYPE1_ALIAS))
 #define GFE_IS_TT(e) ((e)->type == GP_FONT_ENTRY_TRUETYPE)
+#define GFE_IS_S(e) ((e)->type == GP_FONT_ENTRY_SPECIAL)
 
 static GtkObjectClass *parent_class;
 
@@ -365,26 +374,28 @@ gnome_font_face_new (const gchar * name)
 	GPFontMap * map;
 	GPFontEntry * e;
 
-	g_return_val_if_fail (name != NULL, NULL);
+	if (name) {
+		map = gp_fontmap_get ();
 
-	map = gp_fontmap_get ();
+		e = g_hash_table_lookup (map->fontdict, name);
+		if (!e) {
+			gp_fontmap_release (map);
+			return NULL;
+		}
+		if (e->face) {
+			gnome_font_face_ref (e->face);
+			gp_fontmap_release (map);
+			return e->face;
+		}
+
+		gff_face_from_entry (e);
 
-	e = g_hash_table_lookup (map->fontdict, name);
-	if (!e) {
 		gp_fontmap_release (map);
-		return NULL;
+
+		return (e->face);
+	} else {
+		return gnome_font_unsized_closest (NULL, GNOME_FONT_BOOK, FALSE);
 	}
-	if (e->face) {
-		gnome_font_face_ref (e->face);
-		gp_fontmap_release (map);
-		return e->face;
-	}
-
-	gff_face_from_entry (e);
-
-	gp_fontmap_release (map);
-
-	return (e->face);
 }
 
 /* 
@@ -403,33 +414,57 @@ gnome_font_unsized_closest (const char *family_name,
 	int best_dist, dist;
 	GSList * l;
 
-	g_return_val_if_fail (family_name != NULL, NULL);
-
 	/* This should be reimplemented to use the gnome_font_family_hash. */
 
 	map = gp_fontmap_get ();
 
 	best = NULL;
-	best_dist = 1000000;
 	face = NULL;
 
-	for (l = map->fonts; l != NULL; l = l->next) {
-		entry = (GPFontEntry *) l->data;
-		if (!strcasecmp (family_name, entry->familyname)) {
-			dist = abs (weight - entry->Weight) + 100 * (italic != (entry->ItalicAngle != 0));
-			/* Hack to prefer normal to narrow */
-			if (strstr (entry->speciesname, "Narrow")) dist += 6;
-			if (dist < best_dist) {
-				best_dist = dist;
-				best = entry;
+	if (family_name) {
+
+		best_dist = 1000000;
+
+		for (l = map->fonts; l != NULL; l = l->next) {
+			entry = (GPFontEntry *) l->data;
+			if (!strcasecmp (family_name, entry->familyname)) {
+				dist = abs (weight - entry->Weight) + 100 * (italic != (entry->ItalicAngle != 0));
+				/* Hack to prefer normal to narrow */
+				if (strstr (entry->speciesname, "Narrow")) dist += 6;
+				if (dist < best_dist) {
+					best_dist = dist;
+					best = entry;
+				}
 			}
 		}
+	} else {
+		g_warning ("file %s: line %d: No font name specified, using default", __FILE__, __LINE__);
 	}
 
 	if (best) {
 		face = gnome_font_face_new (best->name);
 	} else {
-		face = gnome_font_face_new ("Helvetica");
+		/* No match, try default */
+		guchar *locale, *p;
+		guchar c[128];
+		GPFontEntry *e;
+		locale = setlocale (LC_ALL, NULL);
+		if (!locale) locale = "C";
+		strncpy (c, locale, 127);
+		c[127] = '\0';
+		p = c;
+		while (isalpha (*p) || (*p == '_')) p += 1;
+		if (*p) *p = '\0';
+		e = g_hash_table_lookup (map->defaultsdict, c);
+		if (!e) e = g_hash_table_lookup (map->defaultsdict, "C");
+		if (e) {
+			if (e->face) {
+				gnome_font_face_ref (e->face);
+			} else {
+				gff_face_from_entry (e);
+			}
+			face = e->face;
+		}
 	}
 	if (!face && map->fonts) {
 		/* No face, no helvetica, load whatever font is first */
@@ -450,13 +485,49 @@ gnome_font_unsized_closest (const char *family_name,
 	return face;
 }
 
+#ifdef USE_GNOME_FONT_FACE_EXPERIMENTAL_INTERFACE
+
+GnomeFontFace *
+gnome_font_face_private_from_files (const guchar *filename, gint facenum, const GSList *additional)
+{
+	static gint num = 0;
+	GnomeFontFace *face;
+	GPFontMap *map;
+	GPFontEntry *e;
+	guchar *n, *f, *s;
+
+	num += 1;
+
+	n = g_strdup_printf ("Gnome Print Internal %d", num);
+	f = g_strdup_printf ("Gnome Print");
+	s = g_strdup_printf ("Internal %d", num);
+
+	map = gp_fontmap_get ();
+
+	e = gp_font_entry_from_files (map, n, f, s, TRUE, filename, facenum, additional);
+
+	if (e) {
+		gff_face_from_entry (e);
+		face = e->face;
+		gp_font_entry_unref (e);
+	} else {
+		face = NULL;
+	}
+
+	gp_fontmap_release (map);
+
+	return face;
+}
+
+#endif
+
 GnomeFont *
 gnome_font_face_get_font (const GnomeFontFace * face, gdouble size, gdouble xres, gdouble yres)
 {
 	g_return_val_if_fail (face != NULL, NULL);
 	g_return_val_if_fail (GNOME_IS_FONT_FACE (face), NULL);
 
-	return gnome_font_new (face->entry->name, size);
+	return gnome_font_face_get_font_full ((GnomeFontFace *) face, size, NULL);
 }
 
 GnomeFont *
@@ -465,20 +536,22 @@ gnome_font_face_get_font_default (const GnomeFontFace * face, gdouble size)
 	g_return_val_if_fail (face != NULL, NULL);
 	g_return_val_if_fail (GNOME_IS_FONT_FACE (face), NULL);
 
-	return gnome_font_face_get_font (face, size, 600.0, 600.0);
+	/* fixme: At least try to look like doing something (Lauris) */
+	return gnome_font_face_get_font_full ((GnomeFontFace *) face, size, NULL);
 }
 
 /* fixme: */
 /* Returns the glyph width in 0.001 unit */
 
 gdouble
-gnome_font_face_get_glyph_width (const GnomeFontFace * face, gint glyph)
+gnome_font_face_get_glyph_width (const GnomeFontFace *face, gint glyph)
 {
 	ArtPoint p;
 
 	g_return_val_if_fail (face != NULL, 0.0);
 	g_return_val_if_fail (GNOME_IS_FONT_FACE (face), 0.0);
 
+	p.x = 0;
 	gnome_font_face_get_glyph_stdadvance (face, glyph, &p);
 
 	return p.x;
@@ -505,6 +578,10 @@ gnome_font_face_lookup_default (const GnomeFontFace * face, gint unicode)
 	/* fixme: Nobody should ask mapping of 0 */
 	if (unicode < 1) return 0;
 
+#if 0
+	g_print ("glyph %d lookup %d\n", unicode, (gint) FT_Get_Char_Index (face->ft_face, unicode));
+#endif
+
 	return FT_Get_Char_Index (face->ft_face, unicode);
 }
 
@@ -521,10 +598,12 @@ gff_load (GnomeFontFace *face)
 		g_return_val_if_fail (ft_result == FT_Err_Ok, FALSE);
 	}
 
-	if (face->entry->type == GP_FONT_ENTRY_ALIAS) {
-		entry = ((GPFontEntryAlias *) face->entry)->ref;
-	} else {
-		entry = face->entry;
+	entry = face->entry;
+
+	while (entry && entry->type == GP_FONT_ENTRY_ALIAS) entry = ((GPFontEntryAlias *) face->entry)->ref;
+	if (!entry) {
+		g_warning ("file %s: line %d: face %s: Floating alias list detected", __FILE__, __LINE__, entry->name);
+		return FALSE;
 	}
 
 	if (GFE_IS_T1_A (entry)) {
@@ -535,7 +614,7 @@ gff_load (GnomeFontFace *face)
 		if (t1->afm.name) {
 			ft_result = FT_Attach_File (ft_face, t1->afm.name);
 			if (ft_result != FT_Err_Ok) {
-				g_warning ("file %s: line %d: face %s: Cannot attach file %s", __FILE__, __LINE__, face->entry->name, t1->afm.name);
+				g_warning ("file %s: line %d: face %s: Cannot attach file %s", __FILE__, __LINE__, entry->name, t1->afm.name);
 			}
 		}
 	} else if (GFE_IS_TT (entry)) {
@@ -543,6 +622,19 @@ gff_load (GnomeFontFace *face)
 		tt = (GPFontEntryTT *) entry;
 		ft_result = FT_New_Face (ft_library, tt->ttf.name, tt->facenum, &ft_face);
 		g_return_val_if_fail (ft_result == FT_Err_Ok, FALSE);
+	} else if (GFE_IS_S (entry)) {
+		GPFontEntrySpecial *s;
+		GSList *l;
+		s = (GPFontEntrySpecial *) entry;
+		ft_result = FT_New_Face (ft_library, s->file.name, 0, &ft_face);
+		g_return_val_if_fail (ft_result == FT_Err_Ok, FALSE);
+		for (l = s->additional; l != NULL; l = l->next) {
+			ft_result = FT_Attach_File (ft_face, (gchar *) l->data);
+			if (ft_result != FT_Err_Ok) {
+				g_warning ("file %s: line %d: face %s: Cannot attach file %s",
+					   __FILE__, __LINE__, entry->name, (gchar *) l->data);
+			}
+		}
 	} else {
 		g_assert_not_reached ();
 	}
@@ -554,7 +646,7 @@ gff_load (GnomeFontFace *face)
 
 	ft_result = FT_Select_Charmap (ft_face, ft_encoding_unicode);
 	if (ft_result != FT_Err_Ok) {
-		g_warning ("file %s: line %d: Face %s does not have unicode charmap", __FILE__, __LINE__, face->entry->name);
+		g_warning ("file %s: line %d: Face %s does not have unicode charmap", __FILE__, __LINE__, entry->name);
 	}
 
 	face->num_glyphs = ft_face->num_glyphs;
@@ -903,8 +995,8 @@ gff_face_from_entry (GPFontEntry *e)
 
 /* PSO (PostScriptObject stuff */
 
-static void gff_pso_ensure_buffer_t1 (GFPSObject *pso);
-static void gff_pso_ensure_buffer_tt (GFPSObject *pso);
+static void gff_pso_ensure_buffer_t1 (GFPSObject *pso, GPFontEntryT1 *t1);
+static void gff_pso_ensure_buffer_tt (GFPSObject *pso, GPFontEntryTT *tt);
 static void gff_pso_ensure_buffer_empty (GFPSObject *pso);
 static void gf_pso_sprintf (GFPSObject * pso, const gchar * format, ...);
 static void gf_pso_ensure_space (GFPSObject *pso, gint size);
@@ -975,19 +1067,31 @@ gnome_font_face_pso_free (GFPSObject *pso)
 void
 gnome_font_face_pso_ensure_buffer (GFPSObject *pso)
 {
+	GPFontEntry *entry;
+
 	g_return_if_fail (pso != NULL);
 
-	switch (pso->face->entry->type) {
+	entry = pso->face->entry;
+
+	while (entry && (entry->type == GP_FONT_ENTRY_ALIAS)) entry = ((GPFontEntryAlias *) entry)->ref;
+	if (!entry) {
+		g_warning ("file %s: line %d: Floating alias list detected for %s",
+			   __FILE__, __LINE__, pso->face->entry->name);
+		gff_pso_ensure_buffer_empty (pso);
+		return;
+	}
+
+	switch (entry->type) {
 	case GP_FONT_ENTRY_TYPE1:
 	case GP_FONT_ENTRY_TYPE1_ALIAS:
-		gff_pso_ensure_buffer_t1 (pso);
+		gff_pso_ensure_buffer_t1 (pso, (GPFontEntryT1 *) entry);
 		break;
 	case GP_FONT_ENTRY_TRUETYPE:
-		gff_pso_ensure_buffer_tt (pso);
+		gff_pso_ensure_buffer_tt (pso, (GPFontEntryTT *) entry);
 		break;
 	default:
 		g_warning ("file %s: line %d: Unknown face entry type %s:%d",
-			   __FILE__, __LINE__, pso->face->entry->name, pso->face->entry->type);
+			   __FILE__, __LINE__, entry->name, entry->type);
 		gff_pso_ensure_buffer_empty (pso);
 		break;
 	}
@@ -996,20 +1100,17 @@ gnome_font_face_pso_ensure_buffer (GFPSObject *pso)
 #define INT32_LSB(q) ((q)[0] + ((q)[1] << 8) + ((q)[2] << 16) + ((q)[3] << 24))
 
 static void
-gff_pso_ensure_buffer_t1 (GFPSObject *pso)
+gff_pso_ensure_buffer_t1 (GFPSObject *pso, GPFontEntryT1 *t1)
 {
-	GPFontEntryT1 *t1;
 	struct stat st;
 	guchar * fbuf;
 	gint fh;
 	const gchar *embeddedname;
 
-	t1 = (GPFontEntryT1 *) pso->face->entry;
-
-	if (pso->face->entry->type == GP_FONT_ENTRY_TYPE1) {
-		embeddedname = pso->face->entry->psname;
+	if (t1->entry.type == GP_FONT_ENTRY_TYPE1) {
+		embeddedname = t1->entry.psname;
 	} else {
-		embeddedname = ((GPFontEntryT1Alias *) pso->face->entry)->alias;
+		embeddedname = ((GPFontEntryT1Alias *) t1)->alias;
 	}
 
 	if (!GFF_LOADED (pso->face)) {
@@ -1184,9 +1285,8 @@ gff_pso_ensure_buffer_t1 (GFPSObject *pso)
  */
 
 static void
-gff_pso_ensure_buffer_tt (GFPSObject *pso)
+gff_pso_ensure_buffer_tt (GFPSObject *pso, GPFontEntryTT *tt)
 {
-	GPFontEntryTT *tt;
 	struct stat st;
 	guchar *fbuf;
 	gint fh;
@@ -1196,11 +1296,9 @@ gff_pso_ensure_buffer_tt (GFPSObject *pso)
 	const ArtDRect *bbox;
 	gint i;
 
-	tt = (GPFontEntryTT *) pso->face->entry;
-
 	/* mmap file */
 	if (stat (tt->ttf.name, &st)) {
-		g_warning ("file %s: line %d: Face: %s: Cannot load face", __FILE__, __LINE__, pso->face->entry->name);
+		g_warning ("file %s: line %d: Face: %s: Cannot load face", __FILE__, __LINE__, tt->entry.name);
 		gff_pso_ensure_buffer_empty (pso);
 		return;
 	}
@@ -1220,7 +1318,7 @@ gff_pso_ensure_buffer_tt (GFPSObject *pso)
 		return;
 	}
 
-	embeddedname = pso->face->entry->psname;
+	embeddedname = tt->entry.psname;
 
 	strings = gp_tt_split_file (fbuf, st.st_size);
 	if (strings) {
@@ -1432,92 +1530,6 @@ gf_pso_ensure_space (GFPSObject *pso, gint size)
 	}
 }
 
-
-#if 0
-static int
-read_int32_lsb (const char *p)
-{
-  const unsigned char *q = (unsigned char *)p;
-
-  return q[0] + (q[1] << 8) + (q[2] << 16) + (q[3] << 24);
-}
-
-/* this is actually the same as a pfb to pfa converter
-
- Reference: Adobe technical note 5040, "Supporting Downloadable PostScript
- Language Fonts", page 9 */
-static char *
-pfb_to_flat (const char *input, int input_size)
-{
-	const unsigned char *in = (unsigned char *)input;
-  char *flat;
-  int flat_size, flat_size_max;
-  int in_idx;
-  int length;
-  int i;
-  const char hextab[16] = "0123456789abcdef";
-
-  flat_size = 0;
-  flat_size_max = 32768;
-  flat = g_new (char, flat_size_max);
-
-  for (in_idx = 0; in_idx < input_size;)
-    {
-      if (in[in_idx] != 128)
-	{
-	  g_free (flat);
-	  return NULL;
-	}
-      switch (in[in_idx + 1])
-	{
-	case 1:
-	  length = read_int32_lsb (input + in_idx + 2);
-	  if (flat_size + length > flat_size_max)
-	    {
-	      do
-		flat_size_max <<= 1;
-	      while (flat_size + length > flat_size_max);
-	      flat = g_realloc (flat, flat_size_max);
-	    }
-	  in_idx += 6;
-	  memcpy (flat + flat_size, in + in_idx, length);
-	  flat_size += length;
-	  in_idx += length;
-	  break;
-	case 2:
-	  length = read_int32_lsb (input + in_idx + 2);
-	  if (flat_size + length * 3 > flat_size_max)
-	    {
-	      do
-		flat_size_max <<= 1;
-	      while (flat_size + length * 3 > flat_size_max);
-	      flat = g_realloc (flat, flat_size_max);
-	    }
-	  in_idx += 6;
-	  for (i = 0; i < length; i++)
-	    {
-	      flat[flat_size++] = hextab[in[in_idx] >> 4];
-	      flat[flat_size++] = hextab[in[in_idx] & 15];
-	      in_idx++;
-	      if ((i & 31) == 31 || i == length - 1)
-		flat[flat_size++] = '\n';
-	    }
-	  break;
-	case 3:
-	  /* zero terminate the returned string */
-	  if (flat_size == flat_size_max)
-	    flat = g_realloc (flat, flat_size_max <<= 1);
-	  flat[flat_size] = 0;
-	  return flat;
-	default:
-	    g_free (flat);
-	    return NULL;
-	}
-    }
-  return flat;
-}
-#endif
-
 gchar *
 gnome_font_face_get_pfa (const GnomeFontUnsized *font)
 {
@@ -1606,62 +1618,4 @@ gnome_font_face_get_pfa (const GnomeFontUnsized *font)
 #endif
 }
 
-#if 0
-static gboolean
-gnome_font_face_gt1_load (GnomeFontFace * face)
-{
-	GPFontEntryT1 * t1;
-
-	g_return_val_if_fail ((face->private->entry->type == GP_FONT_ENTRY_TYPE1) ||
-			      (face->private->entry->type == GP_FONT_ENTRY_TYPE1_ALIAS),
-			      FALSE);
-
-	t1 = (GPFontEntryT1 *) face->private->entry;
-
-	if (!face->private->loadedfont) {
-		g_return_val_if_fail (t1->pfb.name != NULL, FALSE);
-		face->private->loadedfont = gt1_load_font (t1->pfb.name, face->private->privencoding);
-		if (!face->private->loadedfont) {
-			face->private->num_glyphs = 1;
-		}
-	}
-
-	return TRUE;
-}
-#endif
-
-#if 0
-static void
-gff_fill_zero_glyph (GnomeFontFace * face)
-{
-	GFFGlyphInfo * info;
-	static ArtBpath undef[] = {{ART_MOVETO, 0.0, 0.0, 0.0, 0.0, 50.0, 50.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 750.0, 50.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 750.0, 950.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 50.0, 950.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 50.0, 50.0},
-				   {ART_MOVETO, 0.0, 0.0, 0.0, 0.0, 100.0, 100.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 700.0, 100.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 700.0, 900.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 100.0, 900.0},
-				   {ART_LINETO, 0.0, 0.0, 0.0, 0.0, 100.0, 100.0},
-				   {ART_END, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0}};
-
-	info = face->private->glyphs;
-
-	info->unicode = 0;
-	info->psname = ".notdef";
-
-	info->advance.x = 800.0;
-	info->advance.y = 0.0;
-
-	info->bbox.x0 = 50.0;
-	info->bbox.y0 = 50.0;
-	info->bbox.x1 = 750.0;
-	info->bbox.y1 = 900.0;
-
-	info->bpath = undef;
-}
-#endif
-
 
