@@ -26,7 +26,13 @@
 #include <glib.h>
 #include <libgnomeprint/gnome-print-encode.h>
 
+#ifdef HAVE_STRING_H
+#include <string.h>
+#endif
+
+#ifdef HAVE_ZLIB_H
 #include "zlib.h"
+#endif
 
 /* TODO : add a parameter to the encodeion methods called "size to beat" so that
    we stop encodeing a row that we know is too large to be chosen. We only need
