@@ -13,6 +13,20 @@
 
 BEGIN_GNOME_DECLS
 
+/*
+ * These are currently defined values for gnome_print_xxx results
+ * feel free to add more, if those make sense
+ */
+
+typedef enum {
+	GNOME_PRINT_OK = 0,
+	GNOME_PRINT_ERROR_UNKNOWN = -1,
+	GNOME_PRINT_ERROR_BADVALUE = -2,
+	GNOME_PRINT_ERROR_NOCURRENTPOINT = -3,
+	GNOME_PRINT_ERROR_NOCURRENTPATH = -4,
+	GNOME_PRINT_ERROR_TEXTCORRUPT = -5,
+}GnomePrintReturnCode;
+
 #define GNOME_TYPE_PRINT_CONTEXT         (gnome_print_context_get_type ())
 #define GNOME_PRINT_CONTEXT(obj)         (GTK_CHECK_CAST ((obj), GNOME_TYPE_PRINT_CONTEXT, GnomePrintContext))
 #define GNOME_PRINT_CONTEXT_CLASS(klass) (GTK_CHECK_CLASS_CAST ((klass), GNOME_TYPE_PRINT_CONTEXT, GnomePrintContextClass))
@@ -66,9 +80,6 @@ int gnome_print_eoclip (GnomePrintContext *pc);
 /* CTM manipulation */
 
 int gnome_print_concat (GnomePrintContext *pc, const double matrix[6]);
-#ifdef SETMATRIX_DEPRECATED
-int gnome_print_setmatrix (GnomePrintContext *pc, const double matrix[6]);
-#endif
 int gnome_print_scale (GnomePrintContext *pc, double sx, double sy);
 int gnome_print_rotate (GnomePrintContext *pc, double theta_in_degrees);
 int gnome_print_translate (GnomePrintContext *pc, double x, double y);
@@ -87,9 +98,12 @@ int gnome_print_stroke (GnomePrintContext *pc);
 /* Text drawing */
 
 int gnome_print_show (GnomePrintContext *pc, char const *text);
-int gnome_print_show_ucs4 (GnomePrintContext * pc, guint32 * buf, gint length);
+int gnome_print_show_sized (GnomePrintContext *pc, const char *text, int bytes);
+/* WARNING: show_ucs4 is DEPRECATED */
+/* WARNING: it uses BIG ENDIAN integers */
+int gnome_print_show_ucs4 (GnomePrintContext *pc, guint32 *buf, gint length);
 int gnome_print_textline (GnomePrintContext *pc, GnomeTextLine *line);
-int gnome_print_glyphlist (GnomePrintContext * pc, GnomeGlyphList * glyphlist);
+int gnome_print_glyphlist (GnomePrintContext *pc, GnomeGlyphList * glyphlist);
 
 /* Images */
 
