@@ -1,11 +1,33 @@
 #define _GP_PS_UNICODE_C_
 
+/*
+ * Unicode to PostScript glyph mapping
+ *
+ * Authors:
+ *   Lauris Kaplinski <lauris@helixcode.com>
+ *
+ * Copyright (C) 1999-2000 Helix Code, Inc.
+ *
+ * TODO:
+ * 1) Use more meaningful data structures instead of hash tables
+ * 2) Really rethink, how that all works wor multinames/multicodes
+ *
+ */
+
+#include <stdlib.h>
+#include <string.h>
+#include <ctype.h>
 #include <libgnomeprint/gp-ps-unicode.h>
 
+/* We maintain separate lookup tables for standard adobe glyph names and for dingbats */
 static GHashTable * ps2uni = NULL;
+static GHashTable * db2uni = NULL;
+/* Unicode to postscript mapping, unified */
 static GHashTable * uni2ps = NULL;
-static GHashTable * ps2ps = NULL;
+/* return GSList of unicode values associated with given ps name */
 static GHashTable * ps2multi = NULL;
+/* The last one simply maps ps name to static ps name */
+static GHashTable * ps2ps = NULL;
 
 static void gen_tables (void);
 
@@ -19,12 +41,44 @@ gp_unicode_from_ps (const gchar * name)
 	return GPOINTER_TO_INT (g_hash_table_lookup (ps2uni, name));
 }
 
+gint
+gp_unicode_from_dingbats (const gchar * name)
+{
+	g_return_val_if_fail (name != NULL, 0);
+
+	if (!db2uni) gen_tables ();
+
+	return GPOINTER_TO_INT (g_hash_table_lookup (db2uni, name));
+}
+
+/*
+ * fixme:
+ *
+ * 1) We do not handle surrogates
+ * 2) Saving const strings to dictionary will eventually eat all our memory
+ *
+ */
+
 const gchar *
 gp_ps_from_unicode (gint unicode)
 {
+	static GHashTable * std2ps = NULL;
+	gchar * ps;
+
 	if (!uni2ps) gen_tables ();
 
-	return g_hash_table_lookup (uni2ps, GINT_TO_POINTER (unicode));
+	ps = g_hash_table_lookup (uni2ps, GINT_TO_POINTER (unicode));
+
+	if (!ps) {
+		if (!std2ps) std2ps = g_hash_table_new (NULL, NULL);
+		ps = g_hash_table_lookup (std2ps, GINT_TO_POINTER (unicode));
+		if (!ps) {
+			ps = g_strdup_printf ("uni%.4X", unicode);
+			g_hash_table_insert (std2ps, GINT_TO_POINTER (unicode), ps);
+		}
+	}
+
+	return ps;
 }
 
 const gchar *
@@ -47,6 +101,133 @@ gp_multi_from_ps (const gchar * name)
 	return g_hash_table_lookup (ps2multi, name);
 }
 
+/* This is UNTESTED */
+
+/* Get ps name semantics according to Adobe guidelines */
+/* Reference: http://partners.adobe.com/asn/developer/typeforum/unicodegn.html */
+
+/* Parses name 'uniABCD' */
+
+static gint gp_univalue (const gchar * n, gint len)
+{
+	/* Step 3 */
+	if (len == 7) {
+		if ((n[0] == 'u') && (n[1] == 'n') && (n[2] == 'i')) {
+			gint i;
+			for (i = 3; i < 7; i++) {
+				if (!isxdigit (n[i])) break;
+				if (n[i] < 'A') break; /* Faster than isupper and should work */
+			}
+			if (i == 7) {
+				gint unicode;
+				unicode = strtol (n + 3, NULL, 16);
+				return unicode;
+			}
+		}
+	}
+
+	return 0;
+}
+
+GSList *
+gp_ps_get_semantics (const gchar * name, gboolean * isDecomp, gboolean * isVar)
+{
+	gchar n[256];
+	gchar * p;
+	GSList * ml;
+	gint unicode;
+	gint len;
+
+	g_return_val_if_fail (name != NULL, NULL);
+	g_return_val_if_fail (isDecomp != NULL, NULL);
+	g_return_val_if_fail (isVar != NULL, NULL);
+
+	*isDecomp = FALSE;
+	*isVar = FALSE;
+
+	strncpy (n, name, 255);
+	n[255] = '\0';
+
+	/* Step 1 */
+	p = strchr (n, '.');
+	if (p) {
+		*isVar = TRUE;
+		*p = '\0';
+		if (n[0] == '\0') return NULL;
+	}
+
+	/* Step 2 */
+	ml = (GSList *) gp_multi_from_ps (n);
+	if (ml) return g_slist_copy (ml);
+	unicode = gp_unicode_from_ps (n);
+	if (unicode > 0) return g_slist_prepend (NULL, GINT_TO_POINTER (unicode));
+
+	/* Step 3 */
+	len = strlen (n);
+	unicode = gp_univalue (n, len);
+	if (unicode) return g_slist_prepend (NULL, GINT_TO_POINTER (unicode));
+
+	/* We can be composite */
+	ml = NULL;
+
+	/* Step 4 */
+	*isDecomp = TRUE;
+	for (p = n; p != NULL; p = strchr (p, '_')) {
+		gchar c[256];
+		gchar * e;
+		gint l;
+		if (*p == '_') p++;
+		e = strchr (p, '_');
+		if (!e) e = p + strlen (p);
+		l = e - p;
+		/* Step 2 with single value */
+		/* fixme: Think out, whether we should somehow handle multiple codepoints */
+		strncpy (c, p, l);
+		c[l] = '\0';
+		unicode = gp_unicode_from_ps (c);
+		if (!unicode) {
+			/* Step 3 */
+			unicode = gp_univalue (c, l);
+		}
+		if (unicode) {
+			ml = g_slist_prepend (ml, GINT_TO_POINTER (unicode));
+		} else {
+			/* Found nonexistent sequence */
+			g_slist_free (ml);
+			ml = NULL;
+			break;
+		}
+	}
+	if (ml) {
+		ml = g_slist_reverse (ml);
+		return ml;
+	}
+
+	/* Step 5 */
+	if ((strncmp (n, "uni", 3) == 0) && ((len & 0x07) == 3)) {
+		gchar c[5];
+		gint i;
+		/* We are candidate for uniABCDEFGH... composite */
+		for (i = 3; i < len; i++) {
+			if (!isxdigit (n[i])) break;
+			if (n[i] < 'A') break; /* Faster than isupper and should work */
+		}
+		if (i == len) {
+			/* We are not valid sequence */
+			c[4] = '\0';
+			for (i = 3; i < len; i+= 4) {
+				strncpy (c, n + i, 4);
+				unicode = strtol (n + 3, NULL, 16);
+				ml = g_slist_prepend (ml, GINT_TO_POINTER (unicode));
+			}
+			ml = g_slist_reverse (ml);
+			return ml;
+		}
+	}
+
+	return NULL;
+}
+
 /*
  * Mapping
  */
@@ -1110,8 +1291,219 @@ static GPPSUniTab unitab[] = {
 { 0x03B6, "zeta" }
 };
 
+/*
+ * fixme: We should really use integer table for dingbats
+ */
+
 #define unitabs (sizeof (unitab) / sizeof (unitab[0]))
 
+static GPPSUniTab dingtab[] = {
+{ 0x0020, "space" }, /* SPACE */
+{ 0x2701, "a1" }, /* UPPER BLADE SCISSORS */
+{ 0x2702, "a2" }, /* BLACK SCISSORS */
+{ 0x2703, "a202" }, /* LOWER BLADE SCISSORS */
+{ 0x2704, "a3" }, /* WHITE SCISSORS */
+{ 0x260E, "a4" }, /* BLACK TELEPHONE */
+{ 0x2706, "a5" }, /* TELEPHONE LOCATION SIGN */
+{ 0x2707, "a119" }, /* TAPE DRIVE */
+{ 0x2708, "a118" }, /* AIRPLANE */
+{ 0x2709, "a117" }, /* ENVELOPE */
+{ 0x261B, "a11" }, /* BLACK RIGHT POINTING INDEX */
+{ 0x261E, "a12" }, /* WHITE RIGHT POINTING INDEX */
+{ 0x270C, "a13" }, /* VICTORY HAND */
+{ 0x270D, "a14" }, /* WRITING HAND */
+{ 0x270E, "a15" }, /* LOWER RIGHT PENCIL */
+{ 0x270F, "a16" }, /* PENCIL */
+{ 0x2710, "a105" }, /* UPPER RIGHT PENCIL */
+{ 0x2711, "a17" }, /* WHITE NIB */
+{ 0x2712, "a18" }, /* BLACK NIB */
+{ 0x2713, "a19" }, /* CHECK MARK */
+{ 0x2714, "a20" }, /* HEAVY CHECK MARK */
+{ 0x2715, "a21" }, /* MULTIPLICATION X */
+{ 0x2716, "a22" }, /* HEAVY MULTIPLICATION X */
+{ 0x2717, "a23" }, /* BALLOT X */
+{ 0x2718, "a24" }, /* HEAVY BALLOT X */
+{ 0x2719, "a25" }, /* OUTLINED GREEK CROSS */
+{ 0x271A, "a26" }, /* HEAVY GREEK CROSS */
+{ 0x271B, "a27" }, /* OPEN CENTRE CROSS */
+{ 0x271C, "a28" }, /* HEAVY OPEN CENTRE CROSS */
+{ 0x271D, "a6" }, /* LATIN CROSS */
+{ 0x271E, "a7" }, /* SHADOWED WHITE LATIN CROSS */
+{ 0x271F, "a8" }, /* OUTLINED LATIN CROSS */
+{ 0x2720, "a9" }, /* MALTESE CROSS */
+{ 0x2721, "a10" }, /* STAR OF DAVID */
+{ 0x2722, "a29" }, /* FOUR TEARDROP-SPOKED ASTERISK */
+{ 0x2723, "a30" }, /* FOUR BALLOON-SPOKED ASTERISK */
+{ 0x2724, "a31" }, /* HEAVY FOUR BALLOON-SPOKED ASTERISK */
+{ 0x2725, "a32" }, /* FOUR CLUB-SPOKED ASTERISK */
+{ 0x2726, "a33" }, /* BLACK FOUR POINTED STAR */
+{ 0x2727, "a34" }, /* WHITE FOUR POINTED STAR */
+{ 0x2605, "a35" }, /* BLACK STAR */
+{ 0x2729, "a36" }, /* STRESS OUTLINED WHITE STAR */
+{ 0x272A, "a37" }, /* CIRCLED WHITE STAR */
+{ 0x272B, "a38" }, /* OPEN CENTRE BLACK STAR */
+{ 0x272C, "a39" }, /* BLACK CENTRE WHITE STAR */
+{ 0x272D, "a40" }, /* OUTLINED BLACK STAR */
+{ 0x272E, "a41" }, /* HEAVY OUTLINED BLACK STAR */
+{ 0x272F, "a42" }, /* PINWHEEL STAR */
+{ 0x2730, "a43" }, /* SHADOWED WHITE STAR */
+{ 0x2731, "a44" }, /* HEAVY ASTERISK */
+{ 0x2732, "a45" }, /* OPEN CENTRE ASTERISK */
+{ 0x2733, "a46" }, /* EIGHT SPOKED ASTERISK */
+{ 0x2734, "a47" }, /* EIGHT POINTED BLACK STAR */
+{ 0x2735, "a48" }, /* EIGHT POINTED PINWHEEL STAR */
+{ 0x2736, "a49" }, /* SIX POINTED BLACK STAR */
+{ 0x2737, "a50" }, /* EIGHT POINTED RECTILINEAR BLACK STAR */
+{ 0x2738, "a51" }, /* HEAVY EIGHT POINTED RECTILINEAR BLACK STAR */
+{ 0x2739, "a52" }, /* TWELVE POINTED BLACK STAR */
+{ 0x273A, "a53" }, /* SIXTEEN POINTED ASTERISK */
+{ 0x273B, "a54" }, /* TEARDROP-SPOKED ASTERISK */
+{ 0x273C, "a55" }, /* OPEN CENTRE TEARDROP-SPOKED ASTERISK */
+{ 0x273D, "a56" }, /* HEAVY TEARDROP-SPOKED ASTERISK */
+{ 0x273E, "a57" }, /* SIX PETALLED BLACK AND WHITE FLORETTE */
+{ 0x273F, "a58" }, /* BLACK FLORETTE */
+{ 0x2740, "a59" }, /* WHITE FLORETTE */
+{ 0x2741, "a60" }, /* EIGHT PETALLED OUTLINED BLACK FLORETTE */
+{ 0x2742, "a61" }, /* CIRCLED OPEN CENTRE EIGHT POINTED STAR */
+{ 0x2743, "a62" }, /* HEAVY TEARDROP-SPOKED PINWHEEL ASTERISK */
+{ 0x2744, "a63" }, /* SNOWFLAKE */
+{ 0x2745, "a64" }, /* TIGHT TRIFOLIATE SNOWFLAKE */
+{ 0x2746, "a65" }, /* HEAVY CHEVRON SNOWFLAKE */
+{ 0x2747, "a66" }, /* SPARKLE */
+{ 0x2748, "a67" }, /* HEAVY SPARKLE */
+{ 0x2749, "a68" }, /* BALLOON-SPOKED ASTERISK */
+{ 0x274A, "a69" }, /* EIGHT TEARDROP-SPOKED PROPELLER ASTERISK */
+{ 0x274B, "a70" }, /* HEAVY EIGHT TEARDROP-SPOKED PROPELLER ASTERISK */
+{ 0x25CF, "a71" }, /* BLACK CIRCLE */
+{ 0x274D, "a72" }, /* SHADOWED WHITE CIRCLE */
+{ 0x25A0, "a73" }, /* BLACK SQUARE */
+{ 0x274F, "a74" }, /* LOWER RIGHT DROP-SHADOWED WHITE SQUARE */
+{ 0x2750, "a203" }, /* UPPER RIGHT DROP-SHADOWED WHITE SQUARE */
+{ 0x2751, "a75" }, /* LOWER RIGHT SHADOWED WHITE SQUARE */
+{ 0x2752, "a204" }, /* UPPER RIGHT SHADOWED WHITE SQUARE */
+{ 0x25B2, "a76" }, /* BLACK UP-POINTING TRIANGLE */
+{ 0x25BC, "a77" }, /* BLACK DOWN-POINTING TRIANGLE */
+{ 0x25C6, "a78" }, /* BLACK DIAMOND */
+{ 0x2756, "a79" }, /* BLACK DIAMOND MINUS WHITE X */
+{ 0x25D7, "a81" }, /* RIGHT HALF BLACK CIRCLE */
+{ 0x2758, "a82" }, /* LIGHT VERTICAL BAR */
+{ 0x2759, "a83" }, /* MEDIUM VERTICAL BAR */
+{ 0x275A, "a84" }, /* HEAVY VERTICAL BAR */
+{ 0x275B, "a97" }, /* HEAVY SINGLE TURNED COMMA QUOTATION MARK ORNAMENT */
+{ 0x275C, "a98" }, /* HEAVY SINGLE COMMA QUOTATION MARK ORNAMENT */
+{ 0x275D, "a99" }, /* HEAVY DOUBLE TURNED COMMA QUOTATION MARK ORNAMENT */
+{ 0x275E, "a100" }, /* HEAVY DOUBLE COMMA QUOTATION MARK ORNAMENT */
+{ 0xF8D7, "a89" }, /* MEDIUM LEFT PARENTHESIS ORNAMENT */
+{ 0xF8D8, "a90" }, /* MEDIUM RIGHT PARENTHESIS ORNAMENT */
+{ 0xF8D9, "a93" }, /* MEDIUM FLATTENED LEFT PARENTHESIS ORNAMENT */
+{ 0xF8DA, "a94" }, /* MEDIUM FLATTENED RIGHT PARENTHESIS ORNAMENT */
+{ 0xF8DB, "a91" }, /* MEDIUM LEFT-POINTING ANGLE BRACKET ORNAMENT */
+{ 0xF8DC, "a92" }, /* MEDIUM RIGHT-POINTING ANGLE BRACKET ORNAMENT */
+{ 0xF8DD, "a205" }, /* HEAVY LEFT-POINTING ANGLE QUOTATION MARK ORNAMENT */
+{ 0xF8DE, "a85" }, /* HEAVY RIGHT-POINTING ANGLE QUOTATION MARK ORNAMENT */
+{ 0xF8DF, "a206" }, /* HEAVY LEFT-POINTING ANGLE BRACKET ORNAMENT */
+{ 0xF8E0, "a86" }, /* HEAVY RIGHT-POINTING ANGLE BRACKET ORNAMENT */
+{ 0xF8E1, "a87" }, /* LIGHT LEFT TORTOISE SHELL BRACKET ORNAMENT */
+{ 0xF8E2, "a88" }, /* LIGHT RIGHT TORTOISE SHELL BRACKET ORNAMENT */
+{ 0xF8E3, "a95" }, /* MEDIUM LEFT CURLY BRACKET ORNAMENT */
+{ 0xF8E4, "a96" }, /* MEDIUM RIGHT CURLY BRACKET ORNAMENT */
+{ 0x2761, "a101" }, /* CURVED STEM PARAGRAPH SIGN ORNAMENT */
+{ 0x2762, "a102" }, /* HEAVY EXCLAMATION MARK ORNAMENT */
+{ 0x2763, "a103" }, /* HEAVY HEART EXCLAMATION MARK ORNAMENT */
+{ 0x2764, "a104" }, /* HEAVY BLACK HEART */
+{ 0x2765, "a106" }, /* ROTATED HEAVY BLACK HEART BULLET */
+{ 0x2766, "a107" }, /* FLORAL HEART */
+{ 0x2767, "a108" }, /* ROTATED FLORAL HEART BULLET */
+{ 0x2663, "a112" }, /* BLACK CLUB SUIT */
+{ 0x2666, "a111" }, /* BLACK DIAMOND SUIT */
+{ 0x2665, "a110" }, /* BLACK HEART SUIT */
+{ 0x2660, "a109" }, /* BLACK SPADE SUIT */
+{ 0x2460, "a120" }, /* CIRCLED DIGIT ONE */
+{ 0x2461, "a121" }, /* CIRCLED DIGIT TWO */
+{ 0x2462, "a122" }, /* CIRCLED DIGIT THREE */
+{ 0x2463, "a123" }, /* CIRCLED DIGIT FOUR */
+{ 0x2464, "a124" }, /* CIRCLED DIGIT FIVE */
+{ 0x2465, "a125" }, /* CIRCLED DIGIT SIX */
+{ 0x2466, "a126" }, /* CIRCLED DIGIT SEVEN */
+{ 0x2467, "a127" }, /* CIRCLED DIGIT EIGHT */
+{ 0x2468, "a128" }, /* CIRCLED DIGIT NINE */
+{ 0x2469, "a129" }, /* CIRCLED NUMBER TEN */
+{ 0x2776, "a130" }, /* DINGBAT NEGATIVE CIRCLED DIGIT ONE */
+{ 0x2777, "a131" }, /* DINGBAT NEGATIVE CIRCLED DIGIT TWO */
+{ 0x2778, "a132" }, /* DINGBAT NEGATIVE CIRCLED DIGIT THREE */
+{ 0x2779, "a133" }, /* DINGBAT NEGATIVE CIRCLED DIGIT FOUR */
+{ 0x277A, "a134" }, /* DINGBAT NEGATIVE CIRCLED DIGIT FIVE */
+{ 0x277B, "a135" }, /* DINGBAT NEGATIVE CIRCLED DIGIT SIX */
+{ 0x277C, "a136" }, /* DINGBAT NEGATIVE CIRCLED DIGIT SEVEN */
+{ 0x277D, "a137" }, /* DINGBAT NEGATIVE CIRCLED DIGIT EIGHT */
+{ 0x277E, "a138" }, /* DINGBAT NEGATIVE CIRCLED DIGIT NINE */
+{ 0x277F, "a139" }, /* DINGBAT NEGATIVE CIRCLED NUMBER TEN */
+{ 0x2780, "a140" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT ONE */
+{ 0x2781, "a141" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT TWO */
+{ 0x2782, "a142" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT THREE */
+{ 0x2783, "a143" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT FOUR */
+{ 0x2784, "a144" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT FIVE */
+{ 0x2785, "a145" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT SIX */
+{ 0x2786, "a146" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT SEVEN */
+{ 0x2787, "a147" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT EIGHT */
+{ 0x2788, "a148" }, /* DINGBAT CIRCLED SANS-SERIF DIGIT NINE */
+{ 0x2789, "a149" }, /* DINGBAT CIRCLED SANS-SERIF NUMBER TEN */
+{ 0x278A, "a150" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT ONE */
+{ 0x278B, "a151" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT TWO */
+{ 0x278C, "a152" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT THREE */
+{ 0x278D, "a153" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FOUR */
+{ 0x278E, "a154" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT FIVE */
+{ 0x278F, "a155" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SIX */
+{ 0x2790, "a156" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT SEVEN */
+{ 0x2791, "a157" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT EIGHT */
+{ 0x2792, "a158" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF DIGIT NINE */
+{ 0x2793, "a159" }, /* DINGBAT NEGATIVE CIRCLED SANS-SERIF NUMBER TEN */
+{ 0x2794, "a160" }, /* HEAVY WIDE-HEADED RIGHTWARDS ARROW */
+{ 0x2192, "a161" }, /* RIGHTWARDS ARROW */
+{ 0x2194, "a163" }, /* LEFT RIGHT ARROW */
+{ 0x2195, "a164" }, /* UP DOWN ARROW */
+{ 0x2798, "a196" }, /* HEAVY SOUTH EAST ARROW */
+{ 0x2799, "a165" }, /* HEAVY RIGHTWARDS ARROW */
+{ 0x279A, "a192" }, /* HEAVY NORTH EAST ARROW */
+{ 0x279B, "a166" }, /* DRAFTING POINT RIGHTWARDS ARROW */
+{ 0x279C, "a167" }, /* HEAVY ROUND-TIPPED RIGHTWARDS ARROW */
+{ 0x279D, "a168" }, /* TRIANGLE-HEADED RIGHTWARDS ARROW */
+{ 0x279E, "a169" }, /* HEAVY TRIANGLE-HEADED RIGHTWARDS ARROW */
+{ 0x279F, "a170" }, /* DASHED TRIANGLE-HEADED RIGHTWARDS ARROW */
+{ 0x27A0, "a171" }, /* HEAVY DASHED TRIANGLE-HEADED RIGHTWARDS ARROW */
+{ 0x27A1, "a172" }, /* BLACK RIGHTWARDS ARROW */
+{ 0x27A2, "a173" }, /* THREE-D TOP-LIGHTED RIGHTWARDS ARROWHEAD */
+{ 0x27A3, "a162" }, /* THREE-D BOTTOM-LIGHTED RIGHTWARDS ARROWHEAD */
+{ 0x27A4, "a174" }, /* BLACK RIGHTWARDS ARROWHEAD */
+{ 0x27A5, "a175" }, /* HEAVY BLACK CURVED DOWNWARDS AND RIGHTWARDS ARROW */
+{ 0x27A6, "a176" }, /* HEAVY BLACK CURVED UPWARDS AND RIGHTWARDS ARROW */
+{ 0x27A7, "a177" }, /* SQUAT BLACK RIGHTWARDS ARROW */
+{ 0x27A8, "a178" }, /* HEAVY CONCAVE-POINTED BLACK RIGHTWARDS ARROW */
+{ 0x27A9, "a179" }, /* RIGHT-SHADED WHITE RIGHTWARDS ARROW */
+{ 0x27AA, "a193" }, /* LEFT-SHADED WHITE RIGHTWARDS ARROW */
+{ 0x27AB, "a180" }, /* BACK-TILTED SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27AC, "a199" }, /* FRONT-TILTED SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27AD, "a181" }, /* HEAVY LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27AE, "a200" }, /* HEAVY UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27AF, "a182" }, /* NOTCHED LOWER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27B1, "a201" }, /* NOTCHED UPPER RIGHT-SHADOWED WHITE RIGHTWARDS ARROW */
+{ 0x27B2, "a183" }, /* CIRCLED HEAVY WHITE RIGHTWARDS ARROW */
+{ 0x27B3, "a184" }, /* WHITE-FEATHERED RIGHTWARDS ARROW */
+{ 0x27B4, "a197" }, /* BLACK-FEATHERED SOUTH EAST ARROW */
+{ 0x27B5, "a185" }, /* BLACK-FEATHERED RIGHTWARDS ARROW */
+{ 0x27B6, "a194" }, /* BLACK-FEATHERED NORTH EAST ARROW */
+{ 0x27B7, "a198" }, /* HEAVY BLACK-FEATHERED SOUTH EAST ARROW */
+{ 0x27B8, "a186" }, /* HEAVY BLACK-FEATHERED RIGHTWARDS ARROW */
+{ 0x27B9, "a195" }, /* HEAVY BLACK-FEATHERED NORTH EAST ARROW */
+{ 0x27BA, "a187" }, /* TEARDROP-BARBED RIGHTWARDS ARROW */
+{ 0x27BB, "a188" }, /* HEAVY TEARDROP-SHANKED RIGHTWARDS ARROW */
+{ 0x27BC, "a189" }, /* WEDGE-TAILED RIGHTWARDS ARROW */
+{ 0x27BD, "a190" }, /* HEAVY WEDGE-TAILED RIGHTWARDS ARROW */
+{ 0x27BE, "a191" } /* OPEN-OUTLINED RIGHTWARDS ARROW */
+};
+
+#define dingtabs (sizeof (dingtab) / sizeof (dingtab[0]))
+
 static void
 gen_tables (void)
 {
@@ -1121,11 +1513,13 @@ gen_tables (void)
 	g_assert (!uni2ps);
 
 	ps2uni = g_hash_table_new (g_str_hash, g_str_equal);
+	db2uni = g_hash_table_new (g_str_hash, g_str_equal);
 	uni2ps = g_hash_table_new (NULL, NULL);
 	ps2ps = g_hash_table_new (g_str_hash, g_str_equal);
 	ps2multi = g_hash_table_new (g_str_hash, g_str_equal);
 
 	g_hash_table_freeze (ps2uni);
+	g_hash_table_freeze (db2uni);
 	g_hash_table_freeze (uni2ps);
 	g_hash_table_freeze (ps2ps);
 	g_hash_table_freeze (ps2multi);
@@ -1146,7 +1540,16 @@ gen_tables (void)
 		g_hash_table_insert (ps2ps, unitab[i].name, unitab[i].name);
 	}
 
+	for (i = 0; i < dingtabs; i++) {
+		g_hash_table_insert (db2uni, dingtab[i].name, GINT_TO_POINTER (dingtab[i].unicode));
+		if (!g_hash_table_lookup (uni2ps, GINT_TO_POINTER (dingtab[i].unicode))) {
+			g_hash_table_insert (uni2ps, GINT_TO_POINTER (dingtab[i].unicode), dingtab[i].name);
+		}
+		g_hash_table_insert (ps2ps, dingtab[i].name, dingtab[i].name);
+	}
+
 	g_hash_table_thaw (ps2uni);
+	g_hash_table_thaw (db2uni);
 	g_hash_table_thaw (uni2ps);
 	g_hash_table_thaw (ps2ps);
 	g_hash_table_thaw (ps2multi);
